@@ -8,7 +8,7 @@
  *
  *    In some countries OX, OX Open-Xchange, open xchange and OXtender
  *    as well as the corresponding Logos OX Open-Xchange and OX are registered
- *    trademarks of the Open-Xchange, Inc. group of companies.
+ *    trademarks of the OX Software GmbH group of companies.
  *    The use of the Logos is not covered by the GNU General Public License.
  *    Instead, you are allowed to use these Logos according to the terms and
  *    conditions of the Creative Commons License, Version 2.5, Attribution,
@@ -28,7 +28,7 @@
  *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
  *    given Attribution for the derivative code and a license granting use.
  *
- *     Copyright (C) 2004-2014 Open-Xchange, Inc.
+ *     Copyright (C) 2016-2020 OX Software GmbH
  *     Mail: info@open-xchange.com
  *
  *
@@ -49,16 +49,27 @@
 
 package com.openexchange.push.imapidle;
 
+import gnu.trove.list.TLongList;
+import gnu.trove.list.array.TLongArrayList;
+import java.util.HashMap;
+import java.util.Map;
 import java.util.concurrent.ExecutionException;
 import java.util.concurrent.Future;
 import java.util.concurrent.TimeUnit;
 import java.util.concurrent.TimeoutException;
 import java.util.concurrent.atomic.AtomicBoolean;
 import java.util.concurrent.locks.LockSupport;
+import javax.mail.FetchProfile;
+import javax.mail.Flags;
 import javax.mail.Folder;
+import javax.mail.Message;
 import javax.mail.MessagingException;
+import javax.mail.UIDFolder;
 import org.slf4j.Logger;
+import com.openexchange.context.ContextService;
 import com.openexchange.exception.OXException;
+import com.openexchange.groupware.contexts.Context;
+import com.openexchange.groupware.ldap.User;
 import com.openexchange.imap.IMAPCapabilities;
 import com.openexchange.imap.IMAPFolderStorage;
 import com.openexchange.imap.IMAPProvider;
@@ -67,14 +78,24 @@ import com.openexchange.mail.api.IMailFolderStorage;
 import com.openexchange.mail.api.IMailFolderStorageDelegator;
 import com.openexchange.mail.api.IMailMessageStorage;
 import com.openexchange.mail.api.MailAccess;
+import com.openexchange.mail.dataobjects.IDMailMessage;
+import com.openexchange.mail.dataobjects.MailMessage;
+import com.openexchange.mail.mime.MessageHeaders;
+import com.openexchange.mail.mime.MimeMailExceptionCode;
+import com.openexchange.mail.mime.MimeTypes;
+import com.openexchange.mail.mime.converters.MimeMessageConverter;
+import com.openexchange.mail.mime.utils.MimeMessageUtility;
 import com.openexchange.mail.service.MailService;
 import com.openexchange.mail.utils.MailFolderUtility;
 import com.openexchange.mailaccount.MailAccount;
 import com.openexchange.mailaccount.MailAccountExceptionCodes;
+import com.openexchange.push.Container;
+import com.openexchange.push.PushEventConstants;
 import com.openexchange.push.PushExceptionCodes;
 import com.openexchange.push.PushListener;
 import com.openexchange.push.PushUtility;
 import com.openexchange.push.imapidle.locking.ImapIdleClusterLock;
+import com.openexchange.push.imapidle.locking.SessionInfo;
 import com.openexchange.server.ServiceExceptionCode;
 import com.openexchange.server.ServiceLookup;
 import com.openexchange.session.Session;
@@ -83,8 +104,12 @@ import com.openexchange.threadpool.ThreadPools;
 import com.openexchange.threadpool.behavior.CallerRunsBehavior;
 import com.openexchange.timer.ScheduledTimerTask;
 import com.openexchange.timer.TimerService;
+import com.openexchange.user.UserService;
+import com.sun.mail.iap.ProtocolException;
 import com.sun.mail.imap.IMAPFolder;
+import com.sun.mail.imap.IMAPMessage;
 import com.sun.mail.imap.IMAPStore;
+import com.sun.mail.imap.protocol.IMAPProtocol;
 
 /**
  * {@link ImapIdlePushListener}
@@ -175,27 +200,58 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
     private final long delayNanos;
     private final PushMode pushMode;
     private final AtomicBoolean canceled;
+    private final boolean permanent;
+    private final boolean supportsPermanentListeners;
     private volatile IMAPFolder imapFolderInUse;
+    private volatile Map<String, Object> additionalProps;
     private volatile long lastLockRefreshNanos;
 
     /**
      * Initializes a new {@link ImapIdlePushListener}.
      */
-    public ImapIdlePushListener(String fullName, int accountId, PushMode pushMode, long delay, Session session, ServiceLookup services) {
+    public ImapIdlePushListener(String fullName, int accountId, PushMode pushMode, long delay, Session session, boolean permanent, boolean supportsPermanentListeners, ServiceLookup services) {
         super();
         canceled = new AtomicBoolean();
+        this.permanent = permanent;
+        this.supportsPermanentListeners = supportsPermanentListeners;
         this.fullName = fullName;
         this.accountId = accountId;
         this.session = session;
         this.delayNanos = TimeUnit.MILLISECONDS.toNanos(delay <= 0 ? 5000L : delay);
         this.services = services;
         this.pushMode = pushMode;
+        additionalProps = null;
         lastLockRefreshNanos = System.nanoTime();
     }
 
+    private boolean isUserValid() {
+        try {
+            ContextService contextService = services.getService(ContextService.class);
+            Context context = contextService.loadContext(session.getContextId());
+            if (!context.isEnabled()) {
+                return false;
+            }
+
+            UserService userService = services.getService(UserService.class);
+            User user = userService.getUser(session.getUserId(), context);
+            return user.isMailEnabled();
+        } catch (OXException e) {
+            return false;
+        }
+    }
+
+    /**
+     * Gets the permanent flag
+     *
+     * @return The permanent flag
+     */
+    public boolean isPermanent() {
+        return permanent;
+    }
+
     @Override
     public void notifyNewMail() throws OXException {
-        PushUtility.triggerOSGiEvent(MailFolderUtility.prepareFullname(accountId, fullName), session);
+        PushUtility.triggerOSGiEvent(MailFolderUtility.prepareFullname(accountId, fullName), session, this.additionalProps, true, true);
     }
 
     @Override
@@ -204,6 +260,11 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
             return;
         }
 
+        if (!isUserValid()) {
+            cancel(false);
+            return;
+        }
+
         String sContextId = Integer.toString(session.getContextId());
         String sUserId = Integer.toString(session.getUserId());
 
@@ -226,18 +287,23 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
                 IMAPStore imapStore = getImapFolderStorageFrom(mailAccess).getImapStore();
                 final IMAPFolder imapFolder = (IMAPFolder) imapStore.getFolder(fullName);
                 this.imapFolderInUse = imapFolder;
+                Map<String, Object> props = new HashMap<String, Object>(3);
+                this.additionalProps = props;
+                long uidNext = -1;
                 try {
+                    // The next expected UID
+                    uidNext = getUIDNext(imapFolder);
+
                     imapFolder.open(Folder.READ_WRITE);
                     LOGGER.debug("Starting IMAP-IDLE run for user {} in context {}.", sUserId, sContextId);
 
                     // Acquire folder counts
+                    int totalCount = imapFolder.getMessageCount();
                     int deletedCount = 0;
-                    int totalCount = 0;
                     final PushMode pushMode = this.pushMode;
                     if (PushMode.ALWAYS == pushMode) {
                         // Operations may be expensive, so only do them in always mode.
                         deletedCount = imapFolder.getDeletedMessageCount();
-                        totalCount = imapFolder.getMessageCount();
                     }
 
                     // Check if canceled meanwhile
@@ -248,7 +314,7 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
 
                     // Refresh lock prior to entering IMAP-IDLE
                     if (doRefreshLock()) {
-                        ImapIdlePushManagerService.getInstance().refreshLock(session);
+                        ImapIdlePushManagerService.getInstance().refreshLock(new SessionInfo(session, permanent));
                     }
 
                     // Are there already new messages?
@@ -295,6 +361,7 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
                             int newMessageCount = imapFolder.getNewMessageCount();
                             if (newMessageCount > 0) {
                                 LOGGER.debug("IMAP-IDLE result for user {} in context {}: Doing push due to {} new mail(s)", sUserId, sContextId, Integer.toString(newMessageCount));
+                                setEventProperties(uidNext, totalCount, props, imapFolder);
                                 notifyNewMail();
                                 notified = true;
                             }
@@ -308,6 +375,7 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
                             int newMessageCount = imapFolder.getNewMessageCount();
                             if (newMessageCount > 0) {
                                 LOGGER.debug("IMAP-IDLE result for user {} in context {}: Doing push due to {} new mail(s)", sUserId, sContextId, Integer.toString(newMessageCount));
+                                setEventProperties(uidNext, totalCount, props, imapFolder);
                                 notifyNewMail();
                                 notified = true;
                                 break;
@@ -319,6 +387,7 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
                             int newDeletedCount = imapFolder.getDeletedMessageCount();
                             if (imapFolder.getDeletedMessageCount() != deletedCount) {
                                 LOGGER.debug("IMAP-IDLE result for user {} in context {}: Doing push due to differing message counts. Current deleted count {} vs. old deleted count {}", sUserId, sContextId, Integer.toString(newDeletedCount), Integer.toString(deletedCount));
+                                setEventProperties(uidNext, totalCount, props, imapFolder);
                                 notifyNewMail();
                                 notified = true;
                                 break;
@@ -327,9 +396,10 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
 
                         // Compare total message counters
                         {
-                            int newTotalCount = imapFolder.getMessageCount();
+                            int newTotalCount = imapFolder.getRealMessageCount();
                             if (newTotalCount != totalCount) {
                                 LOGGER.debug("IMAP-IDLE result for user {} in context {}: Doing push due to differing message counts. Current total count {} vs. old total count {}", sUserId, sContextId, Integer.toString(newTotalCount), Integer.toString(totalCount));
+                                setEventProperties(uidNext, totalCount, props, imapFolder);
                                 notifyNewMail();
                                 notified = true;
                                 break;
@@ -349,6 +419,7 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
                     } catch (final Exception e) {
                         // Ignore
                     }
+                    this.additionalProps = null;
                 }
             } catch (OXException e) {
                 launderOXException(e);
@@ -416,7 +487,7 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
      * @throws InterruptedException If idle'ing thread has been interrupted
      * @throws MessagingException If IMAP-IDLE fails for any reason
      */
-    private boolean doImapIdleTimeoutAware(final IMAPFolder imapFolder) throws InterruptedException, MessagingException {
+    private boolean doImapIdleTimeoutAware(IMAPFolder imapFolder) throws InterruptedException, MessagingException {
         Future<Void> f = ThreadPools.getThreadPool().submit(new ImapIdleTask(imapFolder), CallerRunsBehavior.<Void> getInstance());
         try {
             f.get(TIMEOUT_THRESHOLD_NANOS, TimeUnit.NANOSECONDS);
@@ -464,6 +535,11 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
             if (!capabilities.hasIdle()) {
                 throw PushExceptionCodes.UNEXPECTED_ERROR.create("Primary IMAP account does not support \"IDLE\" capability!");
             }
+        } catch (OXException e) {
+            if (!e.equalsCode(7, "CTX")) {
+                throw e;
+            }
+            // Updating database...
         } finally {
             if (null != access) {
                 access.close(false);
@@ -496,7 +572,7 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
                 }
             }
 
-            // Cancel time task
+            // Cancel timer task
             ScheduledTimerTask timerTask = this.timerTask;
             if (null != timerTask) {
                 this.timerTask = null;
@@ -510,7 +586,7 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
                     // No other listener available
                     // Give up lock and return
                     try {
-                        instance.releaseLock(session);
+                        instance.releaseLock(new SessionInfo(session, permanent));
                     } catch (Exception e) {
                         LOGGER.warn("Failed to release lock for user {} in context {}.", session.getUserId(), session.getContextId(), e);
                     }
@@ -522,7 +598,7 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
                         LOGGER.warn("Failed to start new listener for user {} in context {}.", session.getUserId(), session.getContextId(), e);
                         // Give up lock and return
                         try {
-                            instance.releaseLock(session);
+                            instance.releaseLock(new SessionInfo(session, permanent));
                         } catch (Exception x) {
                             LOGGER.warn("Failed to release DB lock for user {} in context {}.", session.getUserId(), session.getContextId(), x);
                         }
@@ -547,6 +623,184 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
         if ("DBP".equals(e.getPrefix())) {
             throw e;
         }
+        if (MimeMailExceptionCode.LOGIN_FAILED.equals(e)) {
+            Throwable cause = null == e.getCause() ? e : e.getCause();
+            throw PushExceptionCodes.AUTHENTICATION_ERROR.create(cause, new Object[0]);
+        }
+    }
+
+    private long getUIDNext(IMAPFolder imapFolder) throws MessagingException {
+        try {
+            return imapFolder.getUIDNext();
+        } catch (javax.mail.StoreClosedException e) {
+            // Re-throw...
+            throw e;
+        } catch (MessagingException e) {
+            LOGGER.warn("Could not determine UIDNEXT. Assuming -1 instead.", e);
+            return -1L;
+        }
+    }
+
+    private long getUIDNextCustom(IMAPFolder imapFolder) {
+        try {
+            final String fullName = imapFolder.getFullName();
+            return ((Long) imapFolder.doCommand(new IMAPFolder.ProtocolCommand() {
+
+                @Override
+                public Object doCommand(IMAPProtocol p) throws ProtocolException {
+                    String[] item = { "UIDNEXT" };
+                    return Long.valueOf(p.status(fullName, item).uidnext);
+                }
+            })).longValue();
+        } catch (MessagingException e) {
+            LOGGER.warn("Could not determine UIDNEXT. Assuming -1 instead.", e);
+            return -1L;
+        }
+    }
+
+    private void setEventProperties(long uidNext, int totalCount, Map<String, Object> props, IMAPFolder imapFolder) {
+        if (false == supportsPermanentListeners) {
+            return;
+        }
+
+        // Has new messages?
+        if (uidNext > 0 && totalCount >= 0) {
+            try {
+                int newTotalCount = imapFolder.getRealMessageCount();
+                if (newTotalCount > 0 && newTotalCount > totalCount) {
+                    long newUidNext = getUIDNextCustom(imapFolder);
+                    if (newUidNext > 0 && uidNext != newUidNext) {
+                        StringBuilder buf = new StringBuilder(64);
+                        TLongList uids = new TLongArrayList();
+
+                        buf.append(uidNext);
+                        uids.add(uidNext);
+                        for (long uid = uidNext + 1; uid < newUidNext; uid++) {
+                            buf.append(',').append(uid);
+                            uids.add(uid);
+                        }
+                        props.put(PushEventConstants.PROPERTY_IDS, buf.toString());
+
+                        Container<MailMessage> container = fetchMessageInfoFor(uids.toArray(), imapFolder);
+                        if (null != container) {
+                            props.put(PushEventConstants.PROPERTY_CONTAINER, container);
+                        }
+                    }
+                }
+            } catch (MessagingException e) {
+                LOGGER.warn("Could not determine new message count.", e);
+            }
+        }
+
+        // Has deleted messages?
+        if (totalCount > 0) {
+            try {
+                int newTotalCount = imapFolder.getRealMessageCount();
+                if (newTotalCount >= 0 && newTotalCount < totalCount) {
+                    props.put(PushEventConstants.PROPERTY_DELETED, Boolean.TRUE);
+                }
+            } catch (MessagingException e) {
+                LOGGER.warn("Could not determine new message count.", e);
+            }
+        }
+    }
+
+    private static final FetchProfile FETCH_PROFILE_MSG_INFO = new FetchProfile() {
+        {
+            add(FetchProfile.Item.CONTENT_INFO);
+            add(FetchProfile.Item.ENVELOPE);
+            add(FetchProfile.Item.FLAGS);
+            add(FetchProfile.Item.SIZE);
+            add(UIDFolder.FetchProfileItem.UID);
+        }
+    };
+
+    private Container<MailMessage> fetchMessageInfoFor(long[] uids, IMAPFolder imapFolder) throws MessagingException {
+        try {
+            int unread = imapFolder.getUnreadMessageCount();
+            Message[] messages = imapFolder.getMessagesByUID(uids);
+            imapFolder.fetch(messages, FETCH_PROFILE_MSG_INFO);
+
+            String fullName = imapFolder.getFullName();
+            Map<Long, MailMessage> map = new HashMap<Long, MailMessage>(messages.length);
+            for (Message message : messages) {
+                try {
+                    IMAPMessage im = (IMAPMessage) message;
+                    MailMessage mailMessage = convertMessage(im, fullName, unread);
+                    map.put(Long.valueOf(im.getUID()), mailMessage);
+                } catch (Exception e) {
+                    LOGGER.warn("Could not handle message.", e);
+                }
+            }
+
+            Container<MailMessage> container = new Container<MailMessage>();
+            for (long uid : uids) {
+                MailMessage mailMessage = map.get(Long.valueOf(uid));
+                if (null != mailMessage) {
+                    container.add(mailMessage);
+                }
+            }
+            return container;
+        } catch (javax.mail.StoreClosedException e) {
+            // Re-throw...
+            throw e;
+        } catch (MessagingException e) {
+            LOGGER.warn("Could not fetch message info.", e);
+            return null;
+        }
+    }
+
+    private MailMessage convertMessage(IMAPMessage im, String fullName, int unread) throws MessagingException, OXException {
+        MailMessage mailMessage = new IDMailMessage(Long.toString(im.getUID()), fullName);
+        mailMessage.addFrom(MimeMessageConverter.getAddressHeader(MessageHeaders.HDR_FROM, im));
+        mailMessage.addTo(MimeMessageConverter.getAddressHeader(MessageHeaders.HDR_TO, im));
+        mailMessage.addCc(MimeMessageConverter.getAddressHeader(MessageHeaders.HDR_CC, im));
+
+        final Flags msgFlags = im.getFlags();
+        int flags = 0;
+        if (msgFlags.contains(Flags.Flag.ANSWERED)) {
+            flags |= MailMessage.FLAG_ANSWERED;
+        }
+        if (msgFlags.contains(Flags.Flag.DELETED)) {
+            flags |= MailMessage.FLAG_DELETED;
+        }
+        if (msgFlags.contains(Flags.Flag.DRAFT)) {
+            flags |= MailMessage.FLAG_DRAFT;
+        }
+        if (msgFlags.contains(Flags.Flag.FLAGGED)) {
+            flags |= MailMessage.FLAG_FLAGGED;
+        }
+        if (msgFlags.contains(Flags.Flag.RECENT)) {
+            flags |= MailMessage.FLAG_RECENT;
+        }
+        if (msgFlags.contains(Flags.Flag.SEEN)) {
+            flags |= MailMessage.FLAG_SEEN;
+        }
+        if (msgFlags.contains(Flags.Flag.USER)) {
+            flags |= MailMessage.FLAG_USER;
+        }
+
+        mailMessage.setFlags(flags);
+
+        mailMessage.addBcc(MimeMessageConverter.getAddressHeader(MessageHeaders.HDR_BCC, im));
+        {
+            String[] tmp = im.getHeader(MessageHeaders.HDR_CONTENT_TYPE);
+            if ((tmp != null) && (tmp.length > 0)) {
+                mailMessage.setContentType(MimeMessageUtility.decodeMultiEncodedHeader(tmp[0]));
+            } else {
+                mailMessage.setContentType(MimeTypes.MIME_DEFAULT);
+            }
+        }
+        mailMessage.setSentDate(MimeMessageConverter.getSentDate(im));
+        try {
+            mailMessage.setSize(im.getSize());
+        } catch (final Exception e) {
+            // Size unavailable
+            mailMessage.setSize(-1);
+        }
+        mailMessage.setSubject(MimeMessageConverter.getSubject(im));
+        mailMessage.setUnreadMessages(unread);
+        return mailMessage;
     }
 
     private static void closeMailAccess(final MailAccess<?, ?> mailAccess) {
