@@ -8,7 +8,7 @@
  *
  *    In some countries OX, OX Open-Xchange, open xchange and OXtender
  *    as well as the corresponding Logos OX Open-Xchange and OX are registered
- *    trademarks of the Open-Xchange, Inc. group of companies.
+ *    trademarks of the OX Software GmbH group of companies.
  *    The use of the Logos is not covered by the GNU General Public License.
  *    Instead, you are allowed to use these Logos according to the terms and
  *    conditions of the Creative Commons License, Version 2.5, Attribution,
@@ -28,7 +28,7 @@
  *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
  *    given Attribution for the derivative code and a license granting use.
  *
- *     Copyright (C) 2004-2014 Open-Xchange, Inc.
+ *     Copyright (C) 2016-2020 OX Software GmbH
  *     Mail: info@open-xchange.com
  *
  *
@@ -58,7 +58,6 @@ import com.openexchange.exception.OXException;
 import com.openexchange.push.PushExceptionCodes;
 import com.openexchange.server.ServiceExceptionCode;
 import com.openexchange.server.ServiceLookup;
-import com.openexchange.session.Session;
 
 
 /**
@@ -68,7 +67,7 @@ import com.openexchange.session.Session;
  */
 public class HzImapIdleClusterLock extends AbstractImapIdleClusterLock {
 
-    private final String mapName;
+    private volatile String mapName;
     private final AtomicBoolean notActive;
 
     /**
@@ -80,7 +79,20 @@ public class HzImapIdleClusterLock extends AbstractImapIdleClusterLock {
         notActive = new AtomicBoolean();
     }
 
+    /**
+     * Sets the map name
+     *
+     * @param mapName The map name to set
+     */
+    public void setMapName(String mapName) {
+        this.mapName = mapName;
+    }
+
     private IMap<String, String> map(HazelcastInstance hzInstance) throws OXException {
+        String mapName = this.mapName;
+        if (null == mapName) {
+            throw PushExceptionCodes.UNEXPECTED_ERROR.create("Missing map name");
+        }
         try {
             return hzInstance.getMap(mapName);
         } catch (HazelcastInstanceNotActiveException e) {
@@ -97,8 +109,17 @@ public class HzImapIdleClusterLock extends AbstractImapIdleClusterLock {
         notActive.set(true);
     }
 
+    private String generateKey(SessionInfo sessionInfo) {
+        return new StringBuilder(16).append(sessionInfo.getUserId()).append('@').append(sessionInfo.getContextId()).toString();
+    }
+
     @Override
-    public boolean acquireLock(Session session) throws OXException {
+    public Type getType() {
+        return Type.HAZELCAST;
+    }
+
+    @Override
+    public boolean acquireLock(SessionInfo sessionInfo) throws OXException {
         if (notActive.get()) {
             return true;
         }
@@ -107,13 +128,11 @@ public class HzImapIdleClusterLock extends AbstractImapIdleClusterLock {
             throw ServiceExceptionCode.absentService(HazelcastInstance.class);
         }
 
-        int contextId = session.getContextId();
-        int userId = session.getUserId();
-        String key = new StringBuilder(16).append(userId).append('@').append(contextId).toString();
+        String key = generateKey(sessionInfo);
         IMap<String, String> map = map(hzInstance);
 
         long now = System.nanoTime();
-        String previous = map.putIfAbsent(key, generateValue(now, session.getSessionID()));
+        String previous = map.putIfAbsent(key, generateValue(now, sessionInfo));
 
         if (null == previous) {
             // Not present before
@@ -121,17 +140,17 @@ public class HzImapIdleClusterLock extends AbstractImapIdleClusterLock {
         }
 
         // Check if valid
-        if (validValue(previous, now)) {
+        if (validValue(previous, now, hzInstance)) {
             // Locked
             return false;
         }
 
         // Invalid entry - try to replace it mutually exclusive
-        return map.replace(key, previous, generateValue(now, session.getSessionID()));
+        return map.replace(key, previous, generateValue(now, sessionInfo));
     }
 
     @Override
-    public void refreshLock(Session session) throws OXException {
+    public void refreshLock(SessionInfo sessionInfo) throws OXException {
         if (notActive.get()) {
             return;
         }
@@ -140,14 +159,11 @@ public class HzImapIdleClusterLock extends AbstractImapIdleClusterLock {
             throw ServiceExceptionCode.absentService(HazelcastInstance.class);
         }
 
-        int contextId = session.getContextId();
-        int userId = session.getUserId();
-        String key = new StringBuilder(16).append(userId).append('@').append(contextId).toString();
-        map(hzInstance).put(key, generateValue(System.nanoTime(), session.getSessionID()));
+        map(hzInstance).put(generateKey(sessionInfo), generateValue(System.nanoTime(), sessionInfo));
     }
 
     @Override
-    public void releaseLock(Session session) throws OXException {
+    public void releaseLock(SessionInfo sessionInfo) throws OXException {
         if (notActive.get()) {
             return;
         }
@@ -156,10 +172,7 @@ public class HzImapIdleClusterLock extends AbstractImapIdleClusterLock {
             throw ServiceExceptionCode.absentService(HazelcastInstance.class);
         }
 
-        int contextId = session.getContextId();
-        int userId = session.getUserId();
-        String key = new StringBuilder(16).append(userId).append('@').append(contextId).toString();
-        map(hzInstance).remove(key);
+        map(hzInstance).remove(generateKey(sessionInfo));
     }
 
 }
